@@ -11,13 +11,17 @@
 #include <string.h>
 #include <stdlib.h>
 
+#ifndef XC_WIN32
 #include <X11/Intrinsic.h>
 #include <X11/StringDefs.h>
+#endif
 #ifdef TCL_WRAPPER 
 #include <tk.h>
 #else
+#ifndef XC_WIN32
 #include "Xw/Xw.h"
 #endif
+#endif
 
 #include <math.h>
 
@@ -45,6 +49,7 @@ extern XCWindowData *areawin;
 extern char _STR[150];
 extern short fontcount;
 extern fontinfo *fonts;
+extern Boolean was_preselected;
 
 /*---------------------------------------------------------*/
 /* Find the Helvetica font for use in labeling the objects */
@@ -54,6 +59,8 @@ short findhelvetica()
 {
    short fval;
 
+   if (fontcount == 0) loadfontfile("Helvetica");
+
    for (fval = 0; fval < fontcount; fval++)
       if (!strcmp(fonts[fval].psname, "Helvetica"))
 	 break; 
@@ -148,7 +155,7 @@ short pagelinks(int page)
 	 if (xobjs.pagelist[i]->pageinst->thisobject->parts > 0)
 	    if ((i == page) || (xobjs.pagelist[i]->filename &&
 			xobjs.pagelist[page]->filename &&
-			(!strcmp(xobjs.pagelist[i]->filename,
+			(!filecmp(xobjs.pagelist[i]->filename,
 			xobjs.pagelist[page]->filename))))
 	       count++;
 
@@ -164,6 +171,8 @@ short pagelinks(int page)
 /* mode = INDEPENDENT: independent top-level pages	*/
 /* mode = DEPENDENT: dependent pages (subcircuits)	*/
 /* mode = PAGE_DEPEND: subcircuits of the current page,	*/
+/* mode = LINKED_PAGES: subcircuits of the current	*/
+/*	page, including parameter links			*/
 /* mode = TOTAL_PAGES: independent pages + subcircuits  */
 /* mode = ALL_PAGES: all pages in xcircuit		*/
 /*							*/
@@ -180,9 +189,9 @@ short pagelinks(int page)
 short *pagetotals(int page, short mode)
 {
    int i;
-   short *counts, *icount, *result;
+   short *counts, *icount;
 
-   if (xobjs.pagelist[page]->pageinst == NULL) return;
+   if (xobjs.pagelist[page]->pageinst == NULL) return NULL;
 
    counts = (short *)malloc(xobjs.pages * sizeof(short));
    icount = (short *)malloc(xobjs.pages * sizeof(short));
@@ -194,10 +203,12 @@ short *pagetotals(int page, short mode)
    /* Find all the subcircuits of this page */
 
    if (mode != ALL_PAGES)
-      findsubschems(page, xobjs.pagelist[page]->pageinst->thisobject, 0, counts);
+      findsubschems(page, xobjs.pagelist[page]->pageinst->thisobject,
+		0, counts, (mode == LINKED_PAGES) ? TRUE : FALSE);
 
    /* Check independent entries (top-level pages which are not	*/
-   /* subcircuits of another page).  Set the counts entry to -1	*/
+   /* subcircuits of another page, but have the same filename	*/
+   /* as the page we started from).  Set the counts entry to -1	*/
    /* to mark each independent page.				*/
 
    if (mode != PAGE_DEPEND)
@@ -211,7 +222,7 @@ short *pagetotals(int page, short mode)
 	       {
 	          if ((i == page) || (xobjs.pagelist[i]->filename
 			&& xobjs.pagelist[page]->filename
-			&& (!strcmp(xobjs.pagelist[i]->filename,
+			&& (!filecmp(xobjs.pagelist[i]->filename,
 			xobjs.pagelist[page]->filename))))
 		     if ((mode == INDEPENDENT) || (*(counts + i) == 0))
 		        (*(icount + i))++;
@@ -221,11 +232,12 @@ short *pagetotals(int page, short mode)
    /* Check other dependent entries (top-level pages which are 	*/
    /* subcircuits of any independent page).			*/
 
-   if ((mode == DEPENDENT) || (mode == TOTAL_PAGES))
+   if ((mode == DEPENDENT) || (mode == TOTAL_PAGES) || (mode == LINKED_PAGES))
    {
       for (i = 0; i < xobjs.pages; i++)
 	 if ((i != page) && (*(icount + i) > 0))
-	    findsubschems(i, xobjs.pagelist[i]->pageinst->thisobject, 0, counts);
+	    findsubschems(i, xobjs.pagelist[i]->pageinst->thisobject,
+		0, counts, (mode == LINKED_PAGES) ? TRUE : FALSE);
    }
 
    if (mode == INDEPENDENT)
@@ -235,7 +247,8 @@ short *pagetotals(int page, short mode)
    }
    else
    {
-      if (mode == TOTAL_PAGES) {	/* merge dependent and independent */
+      if ((mode == TOTAL_PAGES) || (mode == LINKED_PAGES)) {
+	 /* merge dependent and independent */
 	 for (i = 0; i < xobjs.pages; i++)
 	    if (*(icount + i) > 0)
 	       (*(counts + i))++;
@@ -294,18 +307,32 @@ int is_library(objectptr thisobj)
 }
 
 /*------------------------------------------------------*/
-/* Check if a library (given by name) exists.  If it	*/
-/* does, return the library number.  If not, return -1	*/
+/* Check for library name (string).  Because XCircuit   */
+/* generates the text "Library: <filename>" for		*/
+/* library names, we also check against <filename>	*/
+/* only in these names (this library name syntax is	*/
+/* deprecated, but the check is retained for backwards	*/
+/* compatibility).					*/
+/*							*/
+/* If no library matches the given name, return -1.	*/
 /*------------------------------------------------------*/
 
-int check_library(char *libname)
+int NameToLibrary(char *libname)
 {
+   char *slib;
    int i;
-   
-   for (i = 0; i < xobjs.numlibs; i++)
-      if (!strcmp(xobjs.libtop[i + LIBRARY]->thisobject->name, libname))
-	 return i;
 
+   for (i = 0; i < xobjs.numlibs; i++) {
+      slib = xobjs.libtop[i + LIBRARY]->thisobject->name;
+      if (!strcmp(libname, slib)) {
+	 return i;
+      }
+      else if (!strncmp(slib, "Library: ", 9)) {
+	 if (!strcmp(libname, slib + 9)) {
+	    return i;
+         }
+      }
+   }
    return -1;
 }
 
@@ -507,6 +534,7 @@ objinstptr newpageinst(objectptr pageobj)
    instancedefaults(newinst, pageobj, 0, 0);
    newinst->type = OBJINST;
    newinst->color = DEFAULTCOLOR;
+   newinst->style = NORMAL;	/* Do not scale linewidth with page scale */
    return newinst;
 }
 
@@ -591,6 +619,7 @@ void composepagelib(short mode)
       (*drawbox)->number = 4;
       (*drawbox)->points = (pointlist) malloc(4 * sizeof(XPoint));
       (*drawbox)->passed = NULL;
+      (*drawbox)->cycle = NULL;
       pointptr = (*drawbox)->points;
       pointptr->x = (i % gxsize) * xdel + margin;
       pointptr->y = -(i / gxsize) * ydel - margin;
@@ -603,7 +632,6 @@ void composepagelib(short mode)
       pointptr = (*drawbox)->points + 3;
       pointptr->x = (i % gxsize) * xdel + margin;
       pointptr->y = -((i / gxsize) + 1) * ydel + margin;
-      directory->parts++;
 
       /* each page gets its name printed at the bottom */
 
@@ -620,7 +648,6 @@ void composepagelib(short mode)
 	 strptr->data.string = (char *) malloc(1 + strlen(libobj->name));
 	 strcpy(strptr->data.string, libobj->name);
 	 (*pagelabel)->justify = TOP | NOTBOTTOM | NOTLEFT;
-	 directory->parts++;
       }
    }
 
@@ -757,7 +784,8 @@ void composelib(short mode)
    short llx, lly, width, height;
 
    int totalarea, targetwidth;
-   double scale;
+   double scale, savescale;
+   XPoint savepos;
 
    /* Also make composelib() a wrapper for composepagelib() */
    if ((mode > FONTLIB) && (mode < LIBRARY)) {
@@ -772,6 +800,11 @@ void composelib(short mode)
    for (pgen = libpage->plist; pgen < libpage->plist + libpage->parts; pgen++)
       if (IS_OBJINST(*pgen)) *pgen = NULL;
 
+   /* Before resetting, save the position and scale.  We will restore	*/
+   /* them at the end.							*/
+
+   savepos = libpage->pcorner;
+   savescale = libpage->viewscale;
    reset(libpage, NORMAL);
 
    /* Return if library defines no objects or virtual instances */
@@ -855,9 +888,9 @@ void composelib(short mode)
 
       if (fval < fontcount) {
 	 stringpart *strptr;
+	 /* char *locobjname; (jdk) */
 
 	 NEW_LABEL(drawname, libpage);
-         libpage->parts++;
 	 labeldefaults(*drawname, False, 0, 0);
 	 (*drawname)->color = (spec->virtual) ?
 			OFFBUTTONCOLOR : DEFAULTCOLOR;
@@ -865,6 +898,7 @@ void composelib(short mode)
 	 (*drawname)->string->data.font = fval;
 	 strptr = makesegment(&((*drawname)->string), NULL);
 	 strptr->type = TEXT_STRING;
+
          strptr->data.string = strdup(libobj->name);
          (*drawname)->justify = TOP | NOTBOTTOM | NOTLEFT;
 
@@ -886,6 +920,10 @@ void composelib(short mode)
 
    /* Update the library directory */
    updatepagelib(LIBLIB, mode);
+
+   /* Restore original view position */
+   libpage->pcorner = savepos;
+   libpage->viewscale = savescale;
 }
 
 /*----------------------------------------------------------------*/
@@ -955,6 +993,7 @@ void catvirtualcopy()
       libobj = SELTOOBJINST(newselect);
       libinst = addtoinstlist(i, libobj->thisobject, TRUE);
       instcopy(libinst, libobj);
+      tech_mark_changed(GetObjectTechnology(libobj->thisobject));
    }
 
    clearselects();
@@ -1005,7 +1044,7 @@ void catdelete()
 {
    short *newselect, *libpobjs;
    int i;
-   genericptr *testobj, *tobj;
+   /* genericptr *testobj, *tobj; (jdk) */
    objinstptr libobj;
    liblistptr ilist, llist;
    objectptr *libpage, *compobj, *tlib, *slib;
@@ -1016,6 +1055,8 @@ void catdelete()
       libpage = xobjs.userlibs[i].library;
       libpobjs = &xobjs.userlibs[i].number;
    }
+   else
+      return;  /* To-do: Should have a mechanism here for deleting pages! */
 
    for (newselect = areawin->selectlist; newselect < areawin->selectlist
 	  + areawin->selects; newselect++) {
@@ -1084,13 +1125,16 @@ void catdelete()
 	 }
 
 	 /* Finally, delete the object (permanent---no undoing this!) */
+         tech_mark_changed(GetObjectTechnology(libobj->thisobject));
 	 reset(libobj->thisobject, DESTROY);
       }
    }
 
    clearselects();
 
-   if ((i = is_library(topobject)) >= 0) composelib(LIBRARY + i);
+   if ((i = is_library(topobject)) >= 0) {
+      composelib(LIBRARY + i);
+   }
 
    drawarea(NULL, NULL, NULL);
 }
@@ -1139,7 +1183,7 @@ void linkedlistswap(liblistptr *spec, int o1, int o2)
 
 void linkedlistinsertafter(liblistptr *spec, int o1, int o2)
 {
-   liblistptr s1, s1m, s2, s2m, stmp;
+  liblistptr s1, s1m, s2; /* , s2m, stmp; (jdk) */
    int j;
 
    if ((o1 == o2) || (o1 == (o2 + 1))) return;
@@ -1170,13 +1214,49 @@ void linkedlistinsertafter(liblistptr *spec, int o1, int o2)
    }
 }
 
+/*------------------------------------------------------*/
+/* Set the "changed" flag in a library if any object	*/
+/* in that library has changed.				*/
+/*							*/
+/* If "technology" is NULL, check all objects,		*/
+/* otherwise only check objects with a matching		*/
+/* technology.						*/
+/*------------------------------------------------------*/
+
+void tech_set_changes(TechPtr refns)
+{
+   TechPtr ns;
+   int i, j;
+   objectptr thisobj;
+
+   for (i = 0; i < xobjs.numlibs; i++) {
+      for (j = 0; j < xobjs.userlibs[i].number; j++) {
+         thisobj = *(xobjs.userlibs[i].library + j);
+         if (getchanges(thisobj) > 0) {
+            ns = GetObjectTechnology(thisobj);
+	    if ((refns == NULL) || (refns == ns)) 
+	       ns->flags |= LIBRARY_CHANGED;
+	 }
+      }
+   }
+}
+
+/*------------------------------------------------------*/
+/* Mark a technology as having been modified.		*/
+/*------------------------------------------------------*/
+
+void tech_mark_changed(TechPtr ns)
+{
+   if (ns != NULL) ns->flags |= LIBRARY_CHANGED;
+}
+
 /*------------------------------------------------------*/
 /* Rearrange objects in the library			*/
 /*------------------------------------------------------*/
 
 void catmove(int x, int y)
 {
-   int i, j, k, s1, s2, ocentx, ocenty, rangex, rangey, l;
+  int i, j, k, s1, s2, ocentx, ocenty, rangey, l; /* rangex, (jdk) */
    liblistptr spec;
    objinstptr exchobj, lobj;
 
@@ -1243,10 +1323,15 @@ void catmove(int x, int y)
    for (k = 0; k < areawin->selects; k++) {
 
       /* Find number of lobj (may have changed) */
-      for (s2 = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
+
+      if (lobj == NULL)
+	 s2 = -1;
+      else {
+         for (s2 = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
 		spec = spec->next, s2++)
-	 if (spec->thisinst == lobj)
-	    break;
+	    if (spec->thisinst == lobj)
+	       break;
+      }
 
       exchobj = SELTOOBJINST(areawin->selectlist + k);
       for (s1 = 0, spec = xobjs.userlibs[i].instlist; spec != NULL;
@@ -1265,19 +1350,18 @@ void catmove(int x, int y)
 
    unselect_all();
    composelib(LIBRARY + i);
-   incr_changes(xobjs.libtop[LIBRARY + i]->thisobject);
    if (j != i) {
       composelib(LIBRARY + j);
       centerview(xobjs.libtop[LIBRARY + j]);
-      incr_changes(xobjs.libtop[LIBRARY + j]->thisobject);
    }
 
    drawarea(NULL, NULL, NULL);
 }
 
-/*--------------------------------------------------------*/
-/* Make a duplicate of an object, put in the User Library */
-/*--------------------------------------------------------*/
+/*------------------------------------------------------*/
+/* Make a duplicate of an object, put in the User	*/
+/* Library or the current library (if we're in one).	*/
+/*------------------------------------------------------*/
 
 void copycat()
 {
@@ -1285,7 +1369,10 @@ void copycat()
    objectptr *newobj, *curlib, oldobj;
    objinstptr libobj;
    oparamptr ops, newops;
-   int i, libnum = USERLIB - LIBRARY ;
+   int i, libnum;
+
+   libnum = is_library(topobject);
+   if (libnum < 0) libnum = USERLIB - LIBRARY;  /* default */
 
    for (newselect = areawin->selectlist; newselect < areawin->selectlist
 	  + areawin->selects; newselect++) {
@@ -1366,41 +1453,11 @@ void copycat()
 	 switch(ELEMENTTYPE(*(oldobj->plist + i))) {
 	    case(PATH): {
 	       register pathptr *npath, cpath = TOPATH(oldobj->plist + i);
-	       register genericptr *gpart;
+
 	       NEW_PATH(npath, (*newobj));
-	       (*npath)->style = cpath->style;
-	       (*npath)->width = cpath->width;
-	       (*npath)->parts = 0;
-	       (*npath)->plist = (genericptr *)malloc(cpath->parts *
-			sizeof(genericptr));
-               (*npath)->passed = NULL;
-	       for (gpart = cpath->plist; gpart < cpath->plist + cpath->parts;
-			gpart++) {
-		  switch(ELEMENTTYPE(*gpart)){
-		     case ARC: {
-			arcptr copyarc = TOARC(gpart);
-			arcptr *newarc;
-			NEW_ARC(newarc, (*npath));
-			arccopy(*newarc, copyarc);
-			(*npath)->parts++;
-			} break;
-		     case POLYGON: {
-                        polyptr copypoly = TOPOLY(gpart);
-                        polyptr *newpoly;
-                        NEW_POLY(newpoly, (*npath));
-                        polycopy(*newpoly, copypoly);
-                        (*npath)->parts++;
-                        } break;
-                     case SPLINE: {
-                        splineptr copyspline = TOSPLINE(gpart);
-                        splineptr *newspline;
-                        NEW_SPLINE(newspline, (*npath));
-                        splinecopy(*newspline, copyspline);
-                        (*npath)->parts++;
-                        } break;
-		  }
-	       }	
+	       pathcopy(*npath, cpath);
 	       } break;
+
 	    case(ARC): {
 	       register arcptr *narc, carc = TOARC(oldobj->plist + i);
 
@@ -1409,38 +1466,33 @@ void copycat()
                } break;
       
 	    case(POLYGON): {
-	       register polyptr *npoly, cpoly;
+	       register polyptr *npoly, cpoly = TOPOLY(oldobj->plist + i);
 
-	       cpoly = TOPOLY(oldobj->plist + i);
 	       NEW_POLY(npoly, (*newobj));
 	       polycopy(*npoly, cpoly);
                } break;
 
 	    case(SPLINE): {
-	       register splineptr *nspl, cspl;
+	       register splineptr *nspl, cspl = TOSPLINE(oldobj->plist + i);
 
-	       cspl = TOSPLINE(oldobj->plist + i);
 	       NEW_SPLINE(nspl, (*newobj));
 	       splinecopy(*nspl, cspl);
       	       } break;
 
 	    case(LABEL): {
-	       register labelptr *nlabel, clabel;
+	       register labelptr *nlabel, clabel = TOLABEL(oldobj->plist + i);
 
-	       clabel = TOLABEL(oldobj->plist + i);
 	       NEW_LABEL(nlabel, (*newobj));
 	       labelcopy(*nlabel, clabel);
                } break;
       
 	    case(OBJINST): {
-	       register objinstptr *ninst, cinst;
+	       register objinstptr *ninst, cinst = TOOBJINST(oldobj->plist + i);
 
-	       cinst = TOOBJINST(oldobj->plist + i);
 	       NEW_OBJINST(ninst, (*newobj));
 	       instcopy(*ninst, cinst);
                } break;
          }
-	 (*newobj)->parts++;
       }
    }
 
@@ -1519,6 +1571,7 @@ void catalog_op(int op, int x, int y)
 		  eventmode = NORMAL_MODE;
 	       }
 	       else if ((op == XCF_Library_Pop) || (op == XCF_Library_Copy)) {
+		  int saveselects;
 
 	          /* revert to old page */
 
@@ -1533,7 +1586,10 @@ void catalog_op(int op, int x, int y)
 
 	          snap(x, y, &oldpos);
 
+		  saveselects = areawin->selects;
+		  areawin->selects = 0;
    	          setpage(FALSE);
+		  areawin->selects = saveselects;
 	    
 	          snap(x, y, &areawin->save);
 	          xdiff = areawin->save.x - oldpos.x;
@@ -1544,7 +1600,6 @@ void catalog_op(int op, int x, int y)
 	          for (newselect = areawin->selectlist; newselect <
 		     areawin->selectlist + areawin->selects; newselect++) {
 		     NEW_OBJINST(newobject, topobject);
-		     topobject->parts++;
 		     instcopy(*newobject, TOOBJINST(libpage->plist + *newselect));
 		     /* color should be recast as current color */
 		     (*newobject)->color = areawin->color;
@@ -1562,7 +1617,6 @@ void catalog_op(int op, int x, int y)
 
 	          if (!flag) {
 		     NEW_OBJINST(newobject, topobject);
-                     topobject->parts++;
 		     instcopy(*newobject, *libobj);
 		     (*newobject)->color = areawin->color;
                      (*newobject)->position.x = areawin->save.x;
@@ -1587,7 +1641,9 @@ void catalog_op(int op, int x, int y)
 	          }
 	          else {
                      eventmode = MOVE_MODE;
-	             areawin->editcycle = 1;
+		     was_preselected = FALSE;
+	 	     register_for_undo(XCF_Library_Pop, UNDO_MORE, areawin->topinstance,
+		  		areawin->selectlist, areawin->selects);
 	          }
 #ifdef TCL_WRAPPER
 		  /* fprintf(stderr, "Creating event handler for xctk_drag: "); */
@@ -1595,8 +1651,6 @@ void catalog_op(int op, int x, int y)
                   Tk_CreateEventHandler(areawin->area, PointerMotionMask,
 			(Tk_EventProc *)xctk_drag, NULL);
 #endif
-	 	  register_for_undo(XCF_Library_Pop, UNDO_MORE, areawin->topinstance,
-				areawin->selectlist, areawin->selects);
                   catreturn();
                }
 
@@ -1614,8 +1668,9 @@ void catalog_op(int op, int x, int y)
 		     newselect = allocselect();
 		     *newselect = newinst;
 		     XcSetFunction(GXcopy);
-		     XcSetForeground(SELECTCOLOR);
-		     UDrawObject(*libobj, SINGLE, SELECTCOLOR, NULL);
+		     XcTopSetForeground(SELECTCOLOR);
+		     UDrawObject(*libobj, SINGLE, SELECTCOLOR,
+				xobjs.pagelist[areawin->page]->wirewidth, NULL);
 		  }
 	       }
 	       break;
@@ -1678,19 +1733,22 @@ void startcatalog(xcWidget w, pointertype libmod, caddr_t nulldata)
       XDefineCursor (dpy, areawin->window, DEFAULTCURSOR);
       eventmode = CATALOG_MODE;
    }
-   else if (eventmode != CATMOVE_MODE)
+   else if (eventmode != CATMOVE_MODE) {
+      /* Don't know why I put this here---causes xcircuit to redraw	*/
+      /* the schematic view when switching between library pages.	*/
+      // finish_op(XCF_Cancel, 0, 0);
       eventmode = CATALOG_MODE;
+   }
 
    /* Push the current page onto the push stack, unless	we're going */
    /* to a library from the library directory or vice versa, or	    */
    /* library to library.					    */
 
-
    if (!(((is_library(topobject) >= 0)
 		|| (areawin->topinstance == xobjs.libtop[LIBLIB])
 		|| (areawin->topinstance == xobjs.libtop[PAGELIB]))
 		&& libmod >= PAGELIB)) {
-      push_stack(&areawin->stack, areawin->topinstance);
+      push_stack(&areawin->stack, areawin->topinstance, NULL);
    }
 
    /* set library as new object */
@@ -1698,10 +1756,13 @@ void startcatalog(xcWidget w, pointertype libmod, caddr_t nulldata)
    topobject->viewscale = areawin->vscale;
    topobject->pcorner = areawin->pcorner;
    areawin->topinstance = xobjs.libtop[libmod];
-   setpage(TRUE);
-   invalidate_graphics(topobject);
 
-   transferselects();
+   if (libmod == FONTLIB)
+      setpage(FALSE);
+   else {
+      setpage(TRUE);
+      transferselects();
+   }
 
    /* draw the new screen */
 
